from collections import UserList
__version__ = '0.2.1'

from viewmask import utils  # noqa  # pylint: disable=unused-import


class Annotations(UserList):
    def __init__(self, initlist=()):
        UserList.__init__(self)
        if isinstance(initlist, Annotations):
            # Already deduped, copy without duplicate checking
            self.data[:] = Annotations.data
        else:
            self.extend(initlist)

    @classmethod
    def from_tcga(cls, xml_tree, fit_spline=False):
        """Extract contours from a TCGA XML annotations file.

        Parameters
        ----------
        xml_tree : xml.etree.ElementTree
            The XML tree of the TCGA annotations file.

        Returns
        -------
        contours : viewmask.Annotations
            A list of contours, where each contour is a 1-dimensional NumPy
            array of coordinate pairs, where each coordinate pair is a list
            with exactly 2 integers, representing the X and Y coordinates,
            respectively.
        """
        from numpy import array as to_numpy_array, int32 as npint32
        contours = [
            to_numpy_array(
                [[float(vertex.get("X")), float(vertex.get("Y"))]
                    for vertex in region],
                # np.int32 is necessary for cv2.drawContours
                dtype=npint32)
            for region in xml_tree.getroot().iter("Vertices")
        ]
        return cls(initlist=contours)

    @classmethod
    def from_qpdata(cls, filepath):
        pass

    @classmethod
    def from_asap(cls, xml_tree):
        """Extract contours from a TCGA XML annotations file.

        Parameters
        ----------
        xml_tree : xml.etree.ElementTree
            The XML tree of the ASAP annotations file.

        Returns
        -------
        contours : viewmask.Annotations
            A list of contours, where each contour is a 1-dimensional NumPy
            array of coordinate pairs, where each coordinate pair is a list
            with exactly 2 integers, representing the X and Y coordinates,
            respectively.
        """
        pass

    def export(self, mode):
        """Export the annotations to a usable format.

        Parameters
        ----------
        mode : {'napari', 'opencv'}
            If mode is 'napari', the annotations will be transposed over their
            main diagonal. This is done by reverting the x and y coordinates.
            If mode is 'opencv', the exported annotations will be unchanged.

        Returns
        -------
        contours : list of numpy.ndarray
            A list of contours, where each contour is a list of coordinates,
            where each coordinate is a list with exactly 2 integers,
            representing the X and Y coordinates, respectively.

        Notes
        -----
        The main diagonal is defined as the line that connects the top-left
        corner and the bottom right corner of an image.
        """
        if mode == 'napari':
            from numpy import flip
            return [flip(coordinate_pair) for coordinate_pair in self.data]
        elif mode == 'opencv':
            return self.data

    def as_image(self):
        """Convert an annotations object to an annotation mask.

        Returns
        -------
        rendered_annotations : numpy.ndarray
            A 3-dimensional NumPy array representing the RGB output image.
        """
        import numpy as np
        from cv2 import drawContours, fillPoly
        contours = self.export('opencv')
        x_max = np.amax([x for contour in contours for x, _ in contour])
        y_max = np.amax([y for contour in contours for _, y in contour])
        shape = (y_max, x_max, 3)
        rendered_annotations = np.zeros(shape, dtype=np.uint8)
        drawContours(rendered_annotations, contours, -1, [0, 255, 0])
        for contour in contours:
            fillPoly(
                rendered_annotations,
                np.array([contour], dtype=np.int32),
                [230, 230, 230]
            )
        return rendered_annotations

    def check(self, v):
        from numpy import ndarray
        if not isinstance(v, ndarray):
            raise TypeError(f'item is not of type {ndarray}')

    def __add__(self, other):
        ret = self.__class__(self)  # Alt: self.copy()
        ret.extend(other)
        return ret

    def __setitem__(self, i, v):
        self.check(v)
        self.list[i] = v

    def insert(self, i, v):
        self.check(v)
        super().insert(i, v)

    def extend(self, other):
        for x in other:
            self.append(x)


__all__ = [
    'utils.file_to_dask_array',
    'utils.centers_of_contours',
    'utils.get_stroke_color',
    'utils.mask_to_contours',
    'utils.centers_to_image',
    'Annotations'
]



import cv2
import numpy as np


# https://github.com/jlevy44/PathFlowAI/blob/888f0867eeed4e1265cafed9b9f0f42bebd6a6ae/pathflowai/utils.py#L86-L129
def file_to_dask_array(
    path,
    tile_size=1000,
    overlap=0,
    remove_last=True,
    allow_unknown_chunksizes=False
):
    """Load an image to a dask array.

    Parameters
    ----------
    path : str
        The path to the image file as a string.
    tile_size : int, optional
        Size of chunk to be read in.
    overlap : int, optional
        Do not modify, overlap between neighboring tiles.
    remove_last : bool, optional
        Remove last tile because it has a custom size.
    allow_unknown_chunksizes : bool, optional
        Allow different chunk sizes, more flexible, but slowdown.

    Returns
    -------
    arr : dask.array.Array
        A Dask Array representing the contents of the image file.

    Examples
    --------
    >>> da_img = file_to_dask_array(path)
    >>> npa_img = arr.compute()  # convert the dask array to a numpy array
    >>> pil_img = Image.fromarray(cv2.resize(
    ...     npa_img,
    ...     dsize=(1440, 700),
    ...     interpolation=cv2.INTER_CUBIC
    ... ))
    >>> pil_img.save(test_image_name)
    """
    if path.endswith('.npy'):
        import dask.array as da

        return da.from_array(np.load(path))
    else:
        import openslide

        img = openslide.open_slide(path)
        if isinstance(img, openslide.OpenSlide):
            from openslide import deepzoom
            import dask.array as da
            import dask.delayed

            gen = deepzoom.DeepZoomGenerator(
                img,
                tile_size=tile_size,
                overlap=overlap,
                limit_bounds=True
            )
            max_level = len(gen.level_dimensions) - 1
            n_tiles_x, n_tiles_y = gen.level_tiles[max_level]

            @dask.delayed(pure=True)
            def get_tile(level, column, row):
                tile = gen.get_tile(level, (column, row))  # PIL.Image
                return da.transpose(
                    da.from_array(np.array(tile)),
                    axes=(1, 0, 2)
                )

            sample_tile_shape = get_tile(max_level, 0, 0).shape.compute()
            rows = range(n_tiles_y - (0 if not remove_last else 1))
            cols = range(n_tiles_x - (0 if not remove_last else 1))
            tiles = [da.concatenate(
                [da.from_delayed(
                    get_tile(max_level, col, row),
                    sample_tile_shape,
                    np.uint8
                ) for row in rows],
                allow_unknown_chunksizes=allow_unknown_chunksizes,
                axis=1
            ) for col in cols]
            arr = da.concatenate(
                tiles,
                allow_unknown_chunksizes=allow_unknown_chunksizes
            ).transpose([1, 0, 2])
            return arr
        else:  # img is instance of openslide.ImageSlide
            import dask_image.imread

            return dask_image.imread.imread(path)


def centers_of_contours(contours):
    """Return the centers of a list of OpenCV contours.

    Parameters
    ----------
    contours : list of numpy.ndarray
        A list of cv2 contours; each contour's center will be calculated. Each
        contour is a list with 1 element, which is a list with 2 elements,
        representing the X and Y coordinates as integers, respectively.

    Returns
    -------
    centers : list of tuple of int
        A list of coordinates, where each coordinate is a tuple with exactly 2
        ints, representing the X and Y coordinates, respectively. Each
        coordinate represents the center of the corresponding contour in
        `contours`. In this method, the center is defined as the centroid of
        the contour. If the centroid cannot be calculated, the circumcenter of
        the center is used.
    """
    centers = []
    for contour in contours:
        M = cv2.moments(contour)
        if M["m00"] != 0:
            # centroid of contour:
            center_x = M["m10"] / M["m00"]
            center_y = M["m01"] / M["m00"]
        else:
            # circumcenter of contour:
            (center_x, center_y), _ = cv2.minEnclosingCircle(contour)
        centers.append((int(center_x), int(center_y)))
    return centers


def get_stroke_color(xml_tree):
    """Determine the line color for annotations from a TCGA annotations file.

    Parameters
    ----------
    xml_tree : xml.etree.ElementTree
        The XML tree of the TCGA annotations file.

    Returns
    -------
    line_color : str
        A string representing the proper hex code to use for the stroke color.
        The output is always 6 characters and does not include the hashtag.
    """
    # TODO: verify that the output actually is always 6 characters
    decimal_color = xml_tree.find('./Annotation').attrib['LineColor']
    line_color = hex(int(decimal_color)).replace('0x', '').zfill(6)
    return line_color


def mask_to_contours(mask):
    """Determine the line color for annotations from a TCGA annotations file.

    Parameters
    ----------
    mask : numpy.ndarray
        A NumPy N-dimensional array representing the image. If `mask` has 3
        dimensions, the image is assumed to be RGB, and will be converted to
        grayscale. If `mask` only has 2 dimensions, the image is assumed to be
        grayscale. The values in the input should be in the range [0, 255].

    Returns
    -------
    contours : list of numpy.ndarray
        A list of contours, where each contour is a list of coordinates, where
        each coordinate is a list of a list of X and Y integers.
    """
    import dask.array as da
    if isinstance(mask, da.Array):
        mask = mask.compute()  # convert to numpy array
    if mask.ndim == 3:
        if mask.shape[2] == 3:
            graymask = cv2.cvtColor(mask, cv2.COLOR_RGB2GRAY)
        else:
            raise ValueError("If mask has 3 dimensions, it must be RGB.")
    elif mask.ndim == 2:
        graymask = np.uint8(mask)
    else:
        raise ValueError("mask must be 2-D or 3-D.")

    contours, _ = cv2.findContours(
        graymask,
        cv2.RETR_LIST,
        cv2.CHAIN_APPROX_NONE
    )
    return contours


def centers_to_image(centers, radius=4, write_color=[255, 0, 0], shape=None):
    """Draw coordinates of centers to a static image.

    Parameters
    ----------
    centers : list of tuple of int
        A list of coordinates, where each coordinate is a tuple with exactly 2
        ints, representing the X and Y coordinates, respectively. Each
        coordinate represents the center of the corresponding contour in
        `contours`.
    radius : int, optional
        The radius of each center, defaults to `4`.
    write_color : array_like of int, optional
        The RGB color that should be used to draw the centers, defaults to
        [255, 0, 0], which is red. `write_color` have a length of 3; each
        integer represents red, green, and blue, respectively.
    shape : tuple of int, optional
        The shape of the output image. Defaults to (x_max, y_max, 3), where
        x_max are the maximum x-coordinate and y-coordinate, respectively, of
        the centers.

    Returns
    -------
    rendered_annotations : numpy.ndarray
        An N-dimensional NumPy array representing the RGB output image with the
        shape defined as `shape`.
    """
    if shape is None:
        x_max = np.amax([x for x, _ in centers])
        y_max = np.amax([y for _, y in centers])
        shape = (y_max, x_max, 3)
    rendered_annotations = np.zeros(shape, dtype=np.uint8)
    for center in centers:
        cv2.circle(rendered_annotations, center, radius, write_color, -1)
    return rendered_annotations


def split_dask_array_by_colors(arr):
    # TODO: this is currently used in viewmask. remove this?
    import dask.array as da
    # this is equivalent to cv2.split, but cv2.split is slow.
    # https://opencv24-python-tutorials.readthedocs.io/en/latest/py_tutorials/py_core/py_basic_ops/py_basic_ops.html#:~:text=cv2.split()%20is%20a%20costly%20operation%20(in%20terms%20of%20time),%20so%20only%20use%20it%20if%20necessary.%20Numpy%20indexing%20is%20much%20more%20efficient%20and%20should%20be%20used%20if%20possible.
    # https://answers.opencv.org/question/3754/how-to-split-cv2imread-to-3-separate-mats/?answer=3757#post-id-3757
    channel_r, channel_g, channel_b = arr[:, :, 0], arr[:, :, 1], arr[:, :, 2]
    red_img = da.moveaxis(da.stack([
        channel_r, da.zeros_like(channel_r), da.zeros_like(channel_r)
    ]), 0, -1)
    green_img = da.moveaxis(da.stack([
        da.zeros_like(channel_g), channel_g, da.zeros_like(channel_g)
    ]), 0, -1)
    blue_img = da.moveaxis(da.stack([
        da.zeros_like(channel_b), da.zeros_like(channel_b), channel_b
    ]), 0, -1)
    return red_img, green_img, blue_img


def fit_spline_to_points(points):
    """Fits a B-spline through a sequence of points.

    Parameters
    ----------
    points : numpy.ndarray
        A list of sample vector arrays representing the curve.

    Returns
    -------
    spline_points : numpy.ndarray
        An N-dimensional NumPy array representing the RGB output image with the
        shape defined as `shape`.
    """
    # https://pathflowinterns.slack.com/archives/DTLUTM8NS/p1597243698287400
    # TODO: are the docstrings/types correct?
    from scipy.interpolate import splprep, splev
    tck, u = splprep(points.T, u=None, s=0.0, per=1)
    u_new = np.linspace(u.min(), u.max(), 1000)
    x_new, y_new = splev(u_new, tck, der=0)
    spline_points = np.vstack((x_new, y_new)).T
    return spline_points



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viewmask']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.1.2,<8.0.0',
 'click>=7.1.2,<8.0.0',
 'dask-image>=0.3.0,<0.4.0',
 'napari[pyside2]>=0.3.4,<0.4.0',
 'numpy>=1.18.4,<2.0.0',
 'opencv-python-headless>=4.2.0,<5.0.0',
 'openslide-python>=1.1.1,<2.0.0',
 'scikit-image>=0.17.2,<0.18.0',
 'toolz>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['viewmask = viewmask.cli:cli']}

setup_kwargs = {
    'name': 'viewmask',
    'version': '0.2.1',
    'description': 'A Python package and CLI to view XML annotations and NumPy masks.',
    'long_description': 'viewmask\n========\nA Python package and CLI to view XML annotations and NumPy masks.\n\n|PyPI version fury.io|\n|PyPI downloads|\n|HitCount|\n|Documentation Status|\n|Travis build|\n|PyPI license|\n\n.. |PyPI version fury.io| image:: https://badge.fury.io/py/viewmask.svg\n   :target: https://pypi.python.org/pypi/viewmask/\n\n.. |PyPI downloads| image:: https://img.shields.io/pypi/dm/viewmask\n   :target: https://pypistats.org/packages/viewmask\n\n.. |HitCount| image:: https://hits.dwyl.com/sumanthratna/viewmask.svg\n   :target: https://hits.dwyl.com/sumanthratna/viewmask\n\n.. |Documentation Status| image:: https://readthedocs.org/projects/viewmask/badge/?version=latest\n   :target: https://viewmask.readthedocs.io/?badge=latest\n\n.. |Travis build| image:: https://travis-ci.com/sumanthratna/viewmask.svg?branch=master\n   :target: https://travis-ci.com/sumanthratna/viewmask\n\n.. |PyPI license| image:: https://img.shields.io/pypi/l/viewmask.svg\n   :target: https://pypi.python.org/pypi/viewmask/\n   \nviewmask is meant to be a quick-and-easy tool for visualizing masks and a quick-and-easy library for working with annotations. For a more powerful library, consider `HistomicsTK <https://github.com/DigitalSlideArchive/HistomicsTK>`_ (see `annotations and masks <https://digitalslidearchive.github.io/HistomicsTK/histomicstk.annotations_and_masks.html>`_).\n\nInstallation\n============\n\npip\n------------\n::\n\n python3 -m pip install --upgrade pip\n python3 -m pip install viewmask\n\nor:\n::\n\n python3 -m pip install --upgrade pip\n python3 -m pip install git+git://github.com/sumanthratna/viewmask.git#egg=viewmask\n\nPoetry\n------------\n::\n\n poetry run python -m pip install --upgrade pip\n poetry add viewmask\n\nor:\n::\n\n poetry run python -m pip install --upgrade pip\n poetry add git+https://github.com/sumanthratna/viewmask.git\n',
    'author': 'sumanthratna',
    'author_email': 'sumanthratna@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sumanthratna/viewmask',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)



#!/usr/bin/env python

import click
import napari
import xml.etree.ElementTree as ET
import numpy as np
from viewmask.utils import (
    file_to_dask_array,
    centers_of_contours,
    get_stroke_color,
    mask_to_contours,
    centers_to_image
)
from os.path import splitext


INTERACTIVE_OPTION_HELP = 'If passed, the annotations will be rendered as ' + \
    'napari objects rather than rendered together and displayed as an image.'


@click.group()
@click.version_option()
def cli():
    pass


@cli.command(name='annotations')
@click.argument(
    'annotations',
    type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    '-i',
    '--interactive',
    default=False,
    show_default=True,
    type=bool,
    is_flag=True,
    help=INTERACTIVE_OPTION_HELP
)
def view_annotations(annotations, interactive):
    annotations_ext = splitext(annotations)[1]
    annotations_are_not_xml = annotations_ext != '.xml'
    if annotations_are_not_xml and interactive is True:
        raise ValueError(
            "The interactive flag is only supported with XML annotations.")

    with napari.gui_qt():
        viewer = napari.Viewer()
        if interactive:
            try:
                tree = ET.parse(annotations)
            except ET.ParseError:
                # TODO: don't just do nothing, some XMLs might actually be
                # unparseable
                pass
            from viewmask import Annotations
            annotations = Annotations.from_tcga(tree)
            regions = annotations.export('napari')
            line_color = get_stroke_color(tree)
            viewer.add_shapes(
                regions,
                shape_type='path',
                edge_color=f"#{line_color}"
            )
            viewer.add_points(
                centers_of_contours(regions),
                name='centers'
            )
        else:
            if annotations_ext == '.npy':
                rendered_annotations = np.load(annotations)
            elif annotations_ext == '.png':
                from dask.array import squeeze
                rendered_annotations = squeeze(file_to_dask_array(annotations))
            elif annotations_ext == '.xml':
                try:
                    tree = ET.parse(annotations)
                except ET.ParseError:
                    # TODO: don't just do nothing, some XMLs might actually be
                    # unparseable
                    pass
                annotations = Annotations.from_tcga(tree)
                rendered_annotations = annotations.as_image()
            else:
                # TODO: raise ValueError
                pass
            viewer.add_image(
                rendered_annotations,
                name='annotations',
                blending='additive',
                multiscale=False,
            )
            from cv2 import inRange as select_color, countNonZero
            mask = select_color(  # select pure blue
                rendered_annotations if isinstance(rendered_annotations,
                                                   np.ndarray) \
                else rendered_annotations.compute(),
                np.array([0, 0, 255]),
                np.array([0, 0, 255])
            )
            if countNonZero(mask) == 0:
                # TODO: it'd be nice if we could check this before computing
                # the blue-based nuclei mask. the current flow control seems
                # unintuitive
                mask = rendered_annotations
            centers = centers_of_contours(mask_to_contours(mask))
            viewer.add_image(
                centers_to_image(centers),
                name='centers',
                rgb=True,
                blending='additive',
                multiscale=False,
            )


@cli.command(name='image')
@click.argument('image', type=click.Path(exists=True, dir_okay=False))
def view_image(image):
    da_img = file_to_dask_array(image)
    with napari.gui_qt():
        napari.view_image(da_img, name='image', multiscale=False)


@cli.command(name='overlay')
@click.argument('image', type=click.Path(exists=True, dir_okay=False))
@click.argument('annotations', type=click.Path(exists=True, dir_okay=False))
@click.option(
    '-i',
    '--interactive',
    default=False,
    show_default=True,
    type=bool,
    is_flag=True,
    help=INTERACTIVE_OPTION_HELP
)
def view_overlay(image, annotations, interactive):
    annotations_ext = splitext(annotations)[1]
    annotations_are_not_xml = annotations_ext != '.xml'
    if annotations_are_not_xml and interactive is True:
        raise ValueError(
            "The interactive flag is only supported with XML annotations.")

    da_img = file_to_dask_array(image)

    with napari.gui_qt():
        viewer = napari.Viewer()

        viewer.add_image(
            da_img,
            name='image',
            blending='additive',
            multiscale=False,
        )

        if interactive:
            try:
                tree = ET.parse(annotations)
            except ET.ParseError:
                # TODO: don't just do nothing, some XMLs might actually be
                # unparseable
                pass
            from viewmask import Annotations
            annotations = Annotations.from_tcga(tree)
            regions = annotations.export('napari')
            line_color = get_stroke_color(tree)
            viewer.add_shapes(
                regions,
                shape_type='path',
                edge_color=f"#{line_color}"
            )
            viewer.add_points(
                centers_of_contours(regions),
                name='centers'
            )
        else:
            if annotations_ext == '.npy':
                rendered_annotations = np.load(annotations)
            elif annotations_ext == '.png':
                from dask.array import squeeze
                rendered_annotations = squeeze(file_to_dask_array(annotations))
            elif annotations_ext == '.xml':
                try:
                    tree = ET.parse(annotations)
                except ET.ParseError:
                    # TODO: don't just do nothing, some XMLs might actually be
                    # unparseable
                    pass
                from viewmask import Annotations
                annotations = Annotations.from_tcga(tree)
                rendered_annotations = annotations.as_image()
            else:
                # TODO: raise ValueError
                pass
            viewer.add_image(
                rendered_annotations,
                name='annotations',
                blending='additive',
                multiscale=False,
            )
            from cv2 import inRange as select_color, countNonZero
            mask = select_color(  # select pure blue
                rendered_annotations if isinstance(rendered_annotations,
                                                   np.ndarray) \
                else rendered_annotations.compute(),
                np.array([0, 0, 255]),
                np.array([0, 0, 255])
            )
            if countNonZero(mask) == 0:
                # TODO: it'd be nice if we could check this before computing
                # the blue-based nuclei mask. the current flow control seems
                # unintuitive
                mask = rendered_annotations
            centers = centers_of_contours(mask_to_contours(mask))
            viewer.add_image(
                centers_to_image(centers),
                name='centers',
                rgb=True,
                blending='additive',
                multiscale=False,
            )


if __name__ == '__main__':
    cli()
